@@ -12,9 +12,11 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+import random
+import string
+
 import flask
 from six import iteritems, string_types
-from werkzeug.local import Local
 
 
 DENY = 'DENY'
@@ -42,7 +44,7 @@ GOOGLE_CSP_POLICY = {
     'default-src': '\'self\' *.gstatic.com',
 }
 
-_sentinel = object()
+NONCE_LENGTH = 16
 
 
 class Talisman(object):
@@ -69,6 +71,7 @@ class Talisman(object):
             content_security_policy=DEFAULT_CSP_POLICY,
             content_security_policy_report_uri=None,
             content_security_policy_report_only=False,
+            content_security_policy_nonce_in=None,
             referrer_policy=DEFAULT_REFERRER_POLICY,
             session_cookie_secure=True,
             session_cookie_http_only=True):
@@ -99,6 +102,8 @@ class Talisman(object):
                 as "report-only", which disables the enforcement by the browser
                 and requires a "report-uri" parameter with a backend to receive
                 the POST data
+            content_security_policy_nonce_in: A list of csp sections to include
+                a per-request none value in
             referrer_policy: A string describing the referrer policy for the
                 response.
             session_cookie_secure: Forces the session cookie to only be sent
@@ -136,6 +141,11 @@ class Talisman(object):
                 'Setting content_security_policy_report_only to True also '
                 'requires a URI to be specified in '
                 'content_security_policy_report_uri')
+        self.content_security_policy_nonce_in = (
+            content_security_policy_nonce_in or []
+        )
+
+        app.jinja_env.globals['csp_nonce'] = self._get_nonce
 
         self.referrer_policy = referrer_policy
 
@@ -147,38 +157,25 @@ class Talisman(object):
         self.force_file_save = force_file_save
 
         self.app = app
-        self.local_options = Local()
 
-        app.before_request(self._update_local_options)
         app.before_request(self._force_https)
+        app.before_request(self._make_nonce)
         app.after_request(self._set_response_headers)
 
-    def _update_local_options(self):
-        """Updates view-local options with defaults or specified values."""
+    def _get_local_options(self):
         view_function = flask.current_app.view_functions.get(
-                flask.request.endpoint)
-
+            flask.request.endpoint)
         view_options = getattr(
             view_function, 'talisman_view_options', {})
 
-        force_https = view_options.get('force_https', _sentinel)
-        frame_options = view_options.get('frame_options', _sentinel)
-        frame_options_allow_from = view_options.get(
-            'frame_options_allow_from', _sentinel)
-        content_security_policy = view_options.get(
-            'content_security_policy', _sentinel)
-        setattr(self.local_options, 'force_https',
-                force_https if force_https is not _sentinel
-                else self.force_https)
-        setattr(self.local_options, 'frame_options',
-                frame_options if frame_options is not _sentinel
-                else self.frame_options)
-        setattr(self.local_options, 'frame_options_allow_from',
-                frame_options_allow_from if frame_options_allow_from
-                is not _sentinel else self.frame_options_allow_from)
-        setattr(self.local_options, 'content_security_policy',
-                content_security_policy if content_security_policy
-                is not _sentinel else self.content_security_policy)
+        view_options.setdefault('force_https', self.force_https)
+        view_options.setdefault('frame_options', self.frame_options)
+        view_options.setdefault(
+            'frame_options_allow_from', self.frame_options_allow_from)
+        view_options.setdefault(
+            'content_security_policy', self.content_security_policy)
+
+        return view_options
 
     def _force_https(self):
         """Redirect any non-https requests to https.
@@ -196,7 +193,9 @@ class Talisman(object):
             flask.request.headers.get('X-Forwarded-Proto', 'http') == 'https',
         ]
 
-        if self.local_options.force_https and not any(criteria):
+        local_options = self._get_local_options()
+
+        if local_options['force_https'] and not any(criteria):
             if flask.request.url.startswith('http://'):
                 url = flask.request.url.replace('http://', 'https://', 1)
                 code = 302
@@ -207,41 +206,66 @@ class Talisman(object):
 
     def _set_response_headers(self, response):
         """Applies all configured headers to the given response."""
-        self._set_frame_options_headers(response.headers)
-        self._set_content_security_policy_headers(response.headers)
+        options = self._get_local_options()
+        self._set_frame_options_headers(response.headers, options)
+        self._set_content_security_policy_headers(response.headers, options)
         self._set_hsts_headers(response.headers)
         self._set_referrer_policy_headers(response.headers)
         return response
 
-    def _set_frame_options_headers(self, headers):
-        headers['X-Frame-Options'] = self.local_options.frame_options
+    def _make_nonce(self):
+        local_options = self._get_local_options()
+        if (
+                local_options['content_security_policy'] and
+                self.content_security_policy_nonce_in and
+                not getattr(flask.request, 'csp_nonce', None)):
+            flask.request.csp_nonce = get_random_string(NONCE_LENGTH)
 
-        if self.local_options.frame_options == ALLOW_FROM:
+    def _get_nonce(self):
+        return getattr(flask.request, 'csp_nonce', '')
+
+    def _set_frame_options_headers(self, headers, options):
+        headers['X-Frame-Options'] = options['frame_options']
+
+        if options['frame_options'] == ALLOW_FROM:
             headers['X-Frame-Options'] += " {}".format(
-                self.local_options.frame_options_allow_from)
+                options['frame_options_allow_from'])
 
-    def _set_content_security_policy_headers(self, headers):
+    def _set_content_security_policy_headers(self, headers, options):
         headers['X-XSS-Protection'] = '1; mode=block'
         headers['X-Content-Type-Options'] = 'nosniff'
 
         if self.force_file_save:
             headers['X-Download-Options'] = 'noopen'
 
-        if not self.local_options.content_security_policy:
+        if not options['content_security_policy']:
             return
 
-        policy = self.local_options.content_security_policy
+        policy = options['content_security_policy']
 
-        if not isinstance(policy, string_types):
-            policies = [
-                '{} {}'.format(
-                    k,
-                    ' '.join(v) if not isinstance(v, string_types) else v)
-                for (k, v)
-                in iteritems(policy)
-            ]
+        if isinstance(policy, string_types):
+            # parse the string into a policy dict
+            policy_string = policy
+            policy = {}
 
-            policy = '; '.join(policies)
+            for policy_part in policy_string.split(';'):
+                policy_parts = policy_part.strip().split(' ')
+                policy[policy_parts[0]] = "".join(policy_parts[1:])
+
+        policies = []
+        for section, content in iteritems(policy):
+            if not isinstance(content, string_types):
+                content = ' '.join(content)
+            policy_part = '{} {}'.format(section, content)
+
+            if (
+                    hasattr(flask.request, 'csp_nonce') and
+                    section in self.content_security_policy_nonce_in):
+                policy_part += " 'nonce-{}'".format(flask.request.csp_nonce)
+
+            policies.append(policy_part)
+
+        policy = '; '.join(policies)
 
         if self.content_security_policy_report_uri and \
                 'report-uri' not in policy:
@@ -276,12 +300,7 @@ class Talisman(object):
     def _set_referrer_policy_headers(self, headers):
         headers['Referrer-Policy'] = self.referrer_policy
 
-    def __call__(
-            self,
-            force_https=_sentinel,
-            frame_options=_sentinel,
-            frame_options_allow_from=_sentinel,
-            content_security_policy=_sentinel):
+    def __call__(self, **kwargs):
         """Use talisman as a decorator to configure options for a particular
         view.
 
@@ -303,10 +322,16 @@ class Talisman(object):
                 return 'Embeddable'
         """
         def decorator(f):
-            setattr(f, 'talisman_view_options', dict(
-                force_https=force_https,
-                frame_options=frame_options,
-                frame_options_allow_from=frame_options_allow_from,
-                content_security_policy=content_security_policy))
+            setattr(f, 'talisman_view_options', kwargs)
             return f
         return decorator
+
+
+def get_random_string(length):
+    allowed_chars = (
+        string.ascii_lowercase +
+        string.ascii_uppercase +
+        string.digits)
+    return ''.join(
+        random.SystemRandom().choice(allowed_chars)
+        for _ in range(length))
